from contextlib import suppress
from io import StringIO
from typing import ClassVar, Self

from ruamel.yaml import YAML, YAMLError


# 非线程安全，但在单个事件循环下是协程安全的。
# 如果运行中不进行写入，只存在读取行为，则可以用于多线程
class BriefConfig:
    _key4merge_files: ClassVar[str] = "other_configs_path"

    @classmethod
    def load(cls, config_path: str) -> Self:
        """将配置文件里的参数，赋予单独的变量，方便后面程序调用"""
        configs = cls._load_config(config_path)
        configs["logging"]  # 实例化类的时候，用 configs 这里面提取的值作为参数
        return cls()

    @classmethod
    def _load_config(cls, config_path: str) -> dict:
        """加载一个配置文件，从中取出其他配置文件的路径（文件不存在不报错），最终合并得到一份配置，如果其他配置里也带有更多配置的路径，同样加载"""
        config = cls._load_config_file(config_path)
        cfg_files = config.pop(cls._key4merge_files, None)
        if not cfg_files:
            return config

        other_configs = {}
        for f in cfg_files:
            with suppress(FileNotFoundError):
                other_config = cls._load_config(f)
                cls._update(other_configs, other_config)
        cls._update(other_configs, config)
        return other_configs

    @classmethod
    def _dump(cls, data) -> str:
        yaml = YAML()
        output = StringIO()
        yaml.dump(data, output)
        return output.getvalue()

    @classmethod
    def _update(cls, config: dict, other_config: dict):
        """
        遍历新的配置中每个键值对，如果在当前配置中不存在，就新增；存在，若是不可变类型，就用新的覆盖；
        若是列表，就在原有的追加；若是字典，就递归。
        """
        for key, val in other_config.items():
            if key not in config:
                config[key] = val
                continue
            if isinstance(val, bool | int | float | str):
                config[key] = val
                continue
            if isinstance(val, list):
                config[key].extend(val)
                continue
            if isinstance(val, dict):
                cls._update(config[key], val)
                continue

    @classmethod
    def _load_config_file(cls, f: str) -> dict:
        yaml = YAML()
        try:
            with open(f, encoding="utf-8") as fp:
                return yaml.load(fp)
        except YAMLError as e:
            raise YAMLError(f"The config file '{f}' is illegal as a YAML: {e}") from e



import logging.config
import os
from dataclasses import dataclass
from typing import Self

from briefconf.v0 import BriefConfig

configfile = os.getenv("BRIEF_CONF_CONFIG_FILE", default="tests/my_config.yaml")


@dataclass(frozen=True, slots=True)
class Config(BriefConfig):
    # 用户关心、需要手动配置的参数
    is_production: bool

    redis_ip: str
    redis_port: int
    redis_user: str
    redis_password: str
    queue_name: str

    enabled_sth: list[str]
    valid_choices: dict[str, str]

    # 默认无须用户改动的参数
    data_dir: str

    # 用户不应该考虑，开发者可以改的参数
    some_name: str

    @classmethod
    def load(cls, config_path: str) -> Self:
        configs = cls._load_config(config_path)
        logging.config.dictConfig(configs["logging"])

        redis_ip, redis_port = configs['redis_server'].split(':')

        return cls(
            is_production=configs['is_production'],
            redis_ip=redis_ip,
            redis_port=int(redis_port),
            redis_user=configs['redis_user'],
            redis_password=configs['redis_password'],
            queue_name=configs['queue_name'],

            enabled_sth=configs['enabled_sth'],
            valid_choices=configs['valid_choices'],

            data_dir=configs['data_dir'],

            some_name="abc",
        )


config = Config.load(os.path.abspath(configfile))

__all__ = ["config"]



import logging.config
from dataclasses import dataclass
from typing import ClassVar, Self

from briefconf.v0 import BriefConfig


@dataclass(frozen=True, slots=True)
class CustomConfig(BriefConfig):
    _key4merge_files: ClassVar[str] = "include"

    data_dir: str

    @classmethod
    def load(cls, config_path: str) -> Self:
        configs = cls._load_config(config_path)
        logging.config.dictConfig(configs["logging"])

        return cls(
            data_dir=configs['data_dir'],
        )



"""
对 briefconf 的基本功能进行测试

cd packages/briefconf && uv run -m pytest -s tests/test_basic_func.py; cd -
"""
import pytest

from tests.config_handle import Config
from tests.custom_config_handle import CustomConfig


def test_basic_behavior():
    """两个配置文件没有重复项，看作加载一个配置文件，此时检查值是否和配置中写的一样"""
    config = Config.load("tests/config.example.yaml")
    assert not config.is_production
    assert config.redis_ip == "1.1.1.1"
    assert config.redis_port == 1111
    assert config.redis_user == "default"
    assert config.redis_password == "123abc"
    assert config.queue_name == "my_queue"

    assert config.enabled_sth == ['A', 'B']
    assert config.valid_choices == {"apple": "APPLE", "banana": "BANANA"}

    assert config.data_dir == "config_and_data_files"

def test_merge_behavior():
    """多个配置文件有重复项，此时检查是否按照规则合并"""
    config = Config.load("tests/my_config.yaml")
    # 字符串、数字、布尔是覆盖行为
    assert config.is_production
    assert config.redis_ip == "1.2.3.4"
    assert config.redis_port == 1234
    assert config.redis_user == "vfly2"
    assert config.redis_password == "abcdefg..-"
    assert config.queue_name == "briefconf_queue"

    assert config.enabled_sth == ['A', 'B', 'C', 'D', 'E']
    assert config.valid_choices == {"apple": "APPLE", "banana": "BANANA", "pear": "PEAR"}

    assert config.data_dir == "config_and_data_files"

def test_key4merge_file():
    """自定义键名，要合并的文件的键"""
    with pytest.raises(KeyError):
        config = Config.load("tests/my_custom_config.yaml")

    config = CustomConfig.load("tests/my_custom_config.yaml")
    assert config.data_dir == "config_and_data_files"
